import os
from functools import cached_property
from typing import List

import ffmpeg


class Metadata:
    def __repr__(self):
        fields = [f'{k}:{v}' for k, v in self.__dict__.items() if v is not None and not k.startswith('_')]
        return '{class_name}({fields})'.format(
            class_name=self.__class__.__name__,
            fields=', '.join(fields)
        )


class StreamMetadata(Metadata):
    def __init__(self, index, codec, title=None):
        self.index = index
        self.codec = codec
        self.title = title

    def details(self):
        return f'{self.title} ({self.codec})'


class AudioMetadata(StreamMetadata):
    def __init__(self, channels: int = 2, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.channels = channels

    def details(self):
        match self.channels:
            case 1:
                channels = 'mono'
            case 2:
                channels = 'stereo'
            case 6:
                channels = '5.1'
            case 8:
                channels = '7.1'
            case _:
                channels = str(self.channels)
        return f'{self.title} ({self.codec}/{channels})'


class FileMetadata(Metadata):
    def __init__(self, filepath: str):
        self.filepath = filepath
        self.ext = filepath.lower().split(".")[-1]

    @cached_property
    def ffoutput(self) -> dict:
        return ffmpeg.probe(self.filepath)

    @cached_property
    def duration(self) -> float:
        return float(self.ffoutput['format']['duration'])

    @cached_property
    def audio_streams(self) -> List[AudioMetadata]:
        return [
            AudioMetadata(
                index=f'0:{stream["index"]}',
                codec=stream['codec_name'],
                title=stream.get('tags', {}).get('language', f'Audio (unknown)'),
                channels=stream['channels'],
            )
            for stream in self.ffoutput['streams']
            if stream.get('codec_type') == 'audio'
        ]

    @cached_property
    def video_streams(self) -> List[StreamMetadata]:
        return [
            StreamMetadata(
                index=f'0:{stream["index"]}',
                codec=stream['codec_name'],
                title=stream.get('tags', {}).get('language', f'Video (unknown)'),
            )
            for stream in self.ffoutput['streams']
            if stream.get('codec_type') == 'video'
        ]

    def details(self):
        return '''
        File: {file}
        Video: {video}
        Audio: {audio}
        '''.format(
            file=os.path.basename(self.filepath),
            video=', '.join([vis.details() for vis in self.video_streams]),
            audio=', '.join([aus.details() for aus in self.audio_streams]),
        )



from typing import List


def selector(entries: List[tuple]):
    """Provides a command line interface for selecting from multiple entries
    :param entries: List of Tuples(entry: Any, label: str)
    """
    match len(entries):
        case 0:
            return None
        case 1:
            return entries[0][0]
        case _:
            entry_labels = '\n'.join([
                f'{index}: {entry[1]}'
                for index, entry in enumerate(entries)
            ])
            try:
                return entries[int(input(f'Found multiple entries, please choose: \n{entry_labels}\n'))][0]
            except (ValueError, IndexError):
                print('Invalid answer! Please try again and type the number of your desired answer.')
                return selector(entries)


def format_bytes(size: int | str | None):
    try:
        power = 2**10
        n = int(size)
        if n <= power:
            return f"{n} B"
        for unit in ['KB', 'MB', 'GB']:
            n /= power
            if n < power:
                return f"{n:.1f} {unit}"
        return f"{n:.1f} TB"
    except (ValueError, TypeError):
        return "-"


def simplify_user_agent(ua: str) -> str:
    if not ua or ua == '-':
        return '-'
    
    ua_lower = ua.lower()
    
    if 'crkey' in ua_lower or 'chromecast' in ua_lower:
        return 'Chromecast'
    if 'curl' in ua_lower:
        return 'Curl'
    if 'wget' in ua_lower:
        return 'Wget'
    if 'python' in ua_lower:
        return 'Python'
    
    # Browser detection
    if 'edg' in ua_lower:
        return 'Edge'
    if 'chrome' in ua_lower and 'chromium' not in ua_lower:
        return 'Chrome'
    if 'firefox' in ua_lower:
        return 'Firefox'
    if 'safari' in ua_lower and 'chrome' not in ua_lower:
        return 'Safari'
        
    # Fallback: return the first part (usually Mozilla/5.0) or a shortened version
    return ua.split(' ')[0]



from argparse import ArgumentParser, Namespace
from importlib import metadata
from os import getenv

from .filedata import FileMetadata, AudioMetadata
from .helper import selector
from .tc import TerminalCast, create_tmp_video_file

VERSION = metadata.version('terminalcast')


def select_audio(media_file_data: FileMetadata, args: Namespace) -> AudioMetadata:
    if args.audio_title:
        return [
            stream
            for stream in media_file_data.audio_streams
            if stream.title == args.audio_title
        ][0]
    elif args.non_interactive:
        return media_file_data.audio_streams[0]
    else:
        return selector(entries=[
            (stream, f'Audio: {stream.title}')
            for stream in media_file_data.audio_streams
        ])


def main():
    print(f'Terminalcast Version {VERSION}')

    parser = ArgumentParser(prog='terminalcast', description='Cast local videos to your chromecast')
    parser.add_argument('filepath', help='file path')
    parser.add_argument(
        '--select-ip',
        action='store_true',
        help='Flag to manually select the correct ip where the hosted file should be provided',
    )
    parser.add_argument(
        '--ip',
        help='IP address where hosted file should be provided',
    )
    parser.add_argument(
        '--audio-title',
        help='Title of desired audio stream',
    )
    parser.add_argument(
        '--non-interactive',
        action='store_true',
        help='For scripts and non interactive environments. Disables "select-ip" and auto-select all values not given',
    )
    parser.add_argument(
        '--known-hosts',
        help='Comma separated list of known Chromecast IPs',
        default=getenv('TERMINALCAST_KNOWN_HOSTS', ''),
    )
    args = parser.parse_args()

    print('----- File information -----')
    media_file_data = FileMetadata(filepath=args.filepath)
    print(media_file_data.details())

    audio_stream = select_audio(media_file_data=media_file_data, args=args)
    print(f'Selecting audio stream "{audio_stream.title}"')

    tmp_file_path = ''
    if audio_stream and audio_stream != media_file_data.audio_streams[0]:
        print('Need to create temp file with selected audio track only')
        tmp_file_path = create_tmp_video_file(
            filepath=args.filepath,
            audio_index=audio_stream.index[-1:],
            duration=media_file_data.duration,
        )

    print('----- Create Terminalcast and select IP -----')
    known_hosts = args.known_hosts.split(',') if args.known_hosts else None
    tcast = TerminalCast(
        filepath=tmp_file_path or args.filepath,
        select_ip=args.ip or (args.select_ip and not args.non_interactive),
        known_hosts=known_hosts,
    )
    print(f'IP: {tcast.ip}')

    print('----- Initializing Chromecast -----')
    print(f'Chromecast: {tcast.cast.cast_info.friendly_name}')
    print(f'Status: {tcast.cast.status}')

    print('----- Starting HTTP Server -----')
    tcast.start_server()

    print('----- Start playing video -----\n')
    tcast.play_video()



import os
import re
import socket
import time
from contextlib import closing
from datetime import datetime
from functools import cached_property
from tempfile import mkstemp
from threading import Thread
from typing import List, Callable

import ffmpeg
from bottle import Bottle, static_file, request, response
from pychromecast import Chromecast, get_chromecasts
from pychromecast.controllers.media import MediaController
from tqdm import tqdm
from waitress import serve

from .helper import format_bytes, selector, simplify_user_agent


class TerminalCast:
    def __init__(self, filepath: str, select_ip: str | bool, known_hosts: List[str] = None):
        self.filepath = os.path.abspath(filepath)
        self.select_ip = select_ip
        self.known_hosts = known_hosts
        self.server_thread = None

    @cached_property
    def ip(self) -> str:
        if isinstance(self.select_ip, str):
            return self.select_ip

        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.connect(("8.8.8.8", 53))
            ip_rec = s.getsockname()[0]

        if not self.select_ip:
            return ip_rec

        ip_list = []
        for _ip in socket.gethostbyname_ex(socket.gethostname())[2]:
            if _ip.startswith('127.'):
                continue
            label = f'{_ip} (recommended)' if _ip == ip_rec else _ip
            ip_list.append((_ip, label))

        if ip_list:
            return selector(ip_list)
        else:
            return ip_rec

    @cached_property
    def port(self) -> int:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(('0.0.0.0', 0))
            return s.getsockname()[1]

    @cached_property
    def cast(self) -> Chromecast:
        print('Searching Chromecasts ...')
        chromecasts, browser = get_chromecasts(known_hosts=self.known_hosts)

        chromecast = selector(entries=[
            (cast, f'{cast.cast_info.friendly_name} ({cast.cast_info.host})')
            for cast in chromecasts
        ])

        if chromecast:
            chromecast.wait()
            return chromecast

        raise NoChromecastAvailable('No Chromecast available')

    def start_server(self):
        self.server_thread = Thread(target=self.run_server)
        self.server_thread.start()
        time.sleep(5)

    def get_video_url(self) -> str:
        return f'http://{self.ip}:{self.port}/video'

    def run_server(self):
        print(self.get_video_url())
        run_http_server(filepath=self.filepath, ip=self.ip, port=self.port)

    def play_video(self):
        mc: MediaController = self.cast.media_controller
        mc.play_media(url=self.get_video_url(), content_type='video/mp4')
        mc.block_until_active()
        print(mc.status)


def run_http_server(filepath: str, ip: str, port: int):
    app = Bottle()

    @app.hook('after_request')
    def log_request():
        ts = datetime.now().astimezone().strftime('%d/%b/%Y %H:%M:%S %z')
        length = format_bytes(response.headers.get('Content-Length'))
        addr = request.remote_addr or '-'
        ua = simplify_user_agent(request.headers.get('User-Agent', '-'))

        print(f'[{ts}] {request.method} {request.path} {response.status_code} ({length}) from {addr} ({ua})')

    @app.get('/video')
    def video():
        resp = static_file(filepath, root='/')
        if 'Last-Modified' in resp.headers:
            del resp.headers['Last-Modified']
        resp.headers['Access-Control-Allow-Origin'] = '*'
        resp.headers['Access-Control-Allow-Methods'] = 'GET, HEAD'
        resp.headers['Access-Control-Allow-Headers'] = 'Content-Type'
        return resp

    print('Starting server')
    serve(app, host=ip, port=port, _quiet=True)


def create_tmp_video_file(
    filepath: str,
    audio_index: int,
    duration: float,
    progress_callback: Callable[[float], None] = None
) -> str:
    """
    Create temporary video file with specified audio track only
    :param filepath: file path of original video file
    :param audio_index: stream index of requested audio track
    :param duration: total duration of the video in seconds
    :param progress_callback: function to call with progress percentage
    :return: filename (including path)
    """
    # Prioritize env var, then fall back to smart detection
    temp_dir = os.getenv('TERMINALCAST_TMP_DIR')
    if not temp_dir or not os.path.isdir(temp_dir):
        if os.path.isdir('/dev/shm'):
            temp_dir = '/dev/shm'
        elif os.path.isdir('/var/tmp'):
            temp_dir = '/var/tmp'
        else:
            temp_dir = None  # Use system default

    tmp_file_path = mkstemp(
        suffix='.mp4',
        prefix=f'terminalcast_pid{os.getpid()}_',
        dir=temp_dir
    )[1]
    os.remove(tmp_file_path)

    print(f'Create temporary video file at {tmp_file_path}')

    input_stream = ffmpeg.input(filepath)
    video = input_stream['v']
    audio = input_stream[audio_index]
    
    process = (
        ffmpeg.output(video, audio, tmp_file_path, codec='copy', progress='pipe:1')
        .run_async(pipe_stdout=True, pipe_stderr=True)
    )

    with tqdm(total=100, desc="Converting") as pbar:
        for line in process.stdout:
            line = line.decode('utf-8')
            if 'out_time_ms' in line:
                time_ms = int(line.split('=')[1])
                progress = (time_ms / (duration * 1000000)) * 100
                pbar.n = int(progress)
                pbar.refresh()
                if progress_callback:
                    progress_callback(progress)

    process.wait()
    print(f'Video created')
    return tmp_file_path


class NoChromecastAvailable(Exception):
    pass
